import pytest
from kubernetes_asyncio.client.exceptions import ApiValueError
from prefect_kubernetes.custom_objects import (
    create_namespaced_custom_object,
    delete_namespaced_custom_object,
    get_namespaced_custom_object,
    get_namespaced_custom_object_status,
    list_namespaced_custom_object,
    patch_namespaced_custom_object,
    replace_namespaced_custom_object,
)


async def test_null_body_raises_error(kubernetes_credentials):
    with pytest.raises(ApiValueError):
        await create_namespaced_custom_object.fn(
            group="my-group",
            version="v1",
            plural="ops",
            body=None,
            kubernetes_credentials=kubernetes_credentials,
        )
    with pytest.raises(ApiValueError):
        await patch_namespaced_custom_object.fn(
            group="my-group",
            version="v1",
            plural="ops",
            name="test-name",
            body=None,
            kubernetes_credentials=kubernetes_credentials,
        )
    with pytest.raises(ApiValueError):
        await replace_namespaced_custom_object.fn(
            group="my-group",
            version="v1",
            plural="ops",
            name="test-name",
            body=None,
            kubernetes_credentials=kubernetes_credentials,
        )


async def test_create_namespaced_crd(
    kubernetes_credentials, _mock_api_custom_objects_client
):
    await create_namespaced_custom_object.fn(
        group="my-group",
        version="v1",
        plural="ops",
        body={
            "api": "v1",
            "kind": "op",
            "metadata": {
                "name": "test",
            },
        },
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert (
        _mock_api_custom_objects_client.return_value.create_namespaced_custom_object.call_args[
            1
        ]["a"]
        == "test"
    )

    assert (
        _mock_api_custom_objects_client.return_value.create_namespaced_custom_object.call_args[
            1
        ]["group"]
        == "my-group"
    )
    assert (
        _mock_api_custom_objects_client.return_value.create_namespaced_custom_object.call_args[
            1
        ]["version"]
        == "v1"
    )
    assert (
        _mock_api_custom_objects_client.return_value.create_namespaced_custom_object.call_args[
            1
        ]["plural"]
        == "ops"
    )
    # We can't have models for Custom Resources.
    assert (
        _mock_api_custom_objects_client.return_value.create_namespaced_custom_object.call_args[
            1
        ]["body"]["metadata"]
        == {"name": "test"}
    )


async def test_get_namespaced_custom_object(
    kubernetes_credentials, _mock_api_custom_objects_client
):
    await get_namespaced_custom_object.fn(
        group="my-group",
        version="v1",
        plural="ops",
        name="test-name",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object.call_args[
            1
        ]["a"]
        == "test"
    )

    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object.call_args[
            1
        ]["group"]
        == "my-group"
    )
    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object.call_args[
            1
        ]["version"]
        == "v1"
    )
    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object.call_args[
            1
        ]["plural"]
        == "ops"
    )
    # We can't have models for Custom Resources.
    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object.call_args[
            1
        ]["name"]
        == "test-name"
    )


async def test_get_namespaced_custom_object_status(
    kubernetes_credentials, _mock_api_custom_objects_client
):
    await get_namespaced_custom_object_status.fn(
        group="my-group",
        version="v1",
        plural="ops",
        name="test-name",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object_status.call_args[
            1
        ]["a"]
        == "test"
    )

    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object_status.call_args[
            1
        ]["group"]
        == "my-group"
    )
    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object_status.call_args[
            1
        ]["version"]
        == "v1"
    )
    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object_status.call_args[
            1
        ]["plural"]
        == "ops"
    )
    # We can't have models for Custom Resources.
    assert (
        _mock_api_custom_objects_client.return_value.get_namespaced_custom_object_status.call_args[
            1
        ]["name"]
        == "test-name"
    )


async def test_delete_namespaced_custom_object(
    kubernetes_credentials, _mock_api_custom_objects_client
):
    await delete_namespaced_custom_object.fn(
        group="my-group",
        version="v1",
        plural="ops",
        name="test-name",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert (
        _mock_api_custom_objects_client.return_value.delete_namespaced_custom_object.call_args[
            1
        ]["a"]
        == "test"
    )

    assert (
        _mock_api_custom_objects_client.return_value.delete_namespaced_custom_object.call_args[
            1
        ]["group"]
        == "my-group"
    )
    assert (
        _mock_api_custom_objects_client.return_value.delete_namespaced_custom_object.call_args[
            1
        ]["version"]
        == "v1"
    )
    assert (
        _mock_api_custom_objects_client.return_value.delete_namespaced_custom_object.call_args[
            1
        ]["plural"]
        == "ops"
    )
    # We can't have models for Custom Resources.
    assert (
        _mock_api_custom_objects_client.return_value.delete_namespaced_custom_object.call_args[
            1
        ]["name"]
        == "test-name"
    )


async def test_list_namespaced_custom_object(
    kubernetes_credentials, _mock_api_custom_objects_client
):
    await list_namespaced_custom_object.fn(
        group="my-group",
        version="v1",
        plural="ops",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert (
        _mock_api_custom_objects_client.return_value.list_namespaced_custom_object.call_args[
            1
        ]["a"]
        == "test"
    )

    assert (
        _mock_api_custom_objects_client.return_value.list_namespaced_custom_object.call_args[
            1
        ]["group"]
        == "my-group"
    )
    assert (
        _mock_api_custom_objects_client.return_value.list_namespaced_custom_object.call_args[
            1
        ]["version"]
        == "v1"
    )
    assert (
        _mock_api_custom_objects_client.return_value.list_namespaced_custom_object.call_args[
            1
        ]["plural"]
        == "ops"
    )


async def test_patch_namespaced_custom_object(
    kubernetes_credentials, _mock_api_custom_objects_client
):
    await patch_namespaced_custom_object.fn(
        group="my-group",
        version="v1",
        plural="ops",
        name="test-name",
        body={
            "api": "v1",
            "kind": "op",
            "metadata": {
                "name": "test",
            },
        },
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert (
        _mock_api_custom_objects_client.return_value.patch_namespaced_custom_object.call_args[
            1
        ]["a"]
        == "test"
    )

    assert (
        _mock_api_custom_objects_client.return_value.patch_namespaced_custom_object.call_args[
            1
        ]["group"]
        == "my-group"
    )
    assert (
        _mock_api_custom_objects_client.return_value.patch_namespaced_custom_object.call_args[
            1
        ]["version"]
        == "v1"
    )
    assert (
        _mock_api_custom_objects_client.return_value.patch_namespaced_custom_object.call_args[
            1
        ]["plural"]
        == "ops"
    )
    assert (
        _mock_api_custom_objects_client.return_value.patch_namespaced_custom_object.call_args[
            1
        ]["name"]
        == "test-name"
    )
    assert (
        _mock_api_custom_objects_client.return_value.patch_namespaced_custom_object.call_args[
            1
        ]["body"]["metadata"]
        == {"name": "test"}
    )


async def test_replace_namespaced_custom_object(
    kubernetes_credentials, _mock_api_custom_objects_client
):
    await replace_namespaced_custom_object.fn(
        group="my-group",
        version="v1",
        plural="ops",
        name="test-name",
        body={
            "api": "v1",
            "kind": "op",
            "metadata": {
                "name": "test",
            },
        },
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert (
        _mock_api_custom_objects_client.return_value.replace_namespaced_custom_object.call_args[
            1
        ]["a"]
        == "test"
    )

    assert (
        _mock_api_custom_objects_client.return_value.replace_namespaced_custom_object.call_args[
            1
        ]["group"]
        == "my-group"
    )
    assert (
        _mock_api_custom_objects_client.return_value.replace_namespaced_custom_object.call_args[
            1
        ]["version"]
        == "v1"
    )
    assert (
        _mock_api_custom_objects_client.return_value.replace_namespaced_custom_object.call_args[
            1
        ]["plural"]
        == "ops"
    )
    assert (
        _mock_api_custom_objects_client.return_value.replace_namespaced_custom_object.call_args[
            1
        ]["name"]
        == "test-name"
    )
    assert (
        _mock_api_custom_objects_client.return_value.replace_namespaced_custom_object.call_args[
            1
        ]["body"]["metadata"]
        == {"name": "test"}
    )



import asyncio
import os
import subprocess
from typing import Any

import anyio
import pytest
from rich.console import Console

from prefect import get_client
from prefect.states import StateType
from prefect_kubernetes_integration_tests.utils import display, k8s, prefect_core

console = Console()

DEFAULT_JOB_VARIABLES: dict[str, Any] = {
    "image": "prefecthq/prefect:3.2.11-python3.12",
}
if os.environ.get("CI", False):
    DEFAULT_JOB_VARIABLES["env"] = {"PREFECT_API_URL": "http://172.17.0.1:4200/api"}

DEFAULT_PARAMETERS = {"n": 5}
# Default source is a simple flow that sleeps
DEFAULT_FLOW_SOURCE = "https://gist.github.com/772d095672484b76da40a4e6158187f0.git"
DEFAULT_FLOW_ENTRYPOINT = "sleeping.py:sleepy"


@pytest.mark.usefixtures("kind_cluster")
async def test_failed_pod_start(
    work_pool_name: str,
):
    """Test flow runs with pods that fail to start are marked as crashed."""
    flow_run = await prefect_core.create_flow_run(
        source=DEFAULT_FLOW_SOURCE,
        entrypoint=DEFAULT_FLOW_ENTRYPOINT,
        name="failed-pod-start",
        work_pool_name=work_pool_name,
        # Use an invalid image to ensure the pod will fail to start
        job_variables=DEFAULT_JOB_VARIABLES
        | {"image": "ceci-nest-pas-une-image:latest"},
        parameters=DEFAULT_PARAMETERS,
        flow_run_name="failed-pod-start-test",
    )

    display.print_flow_run_created(flow_run)

    # Use run_once mode for the crash detection test since we just need one attempt
    # Unlike the subprocess approach, this will block until the worker completes
    print("Starting worker in run_once mode to detect crash...")
    prefect_core.start_worker(work_pool_name, run_once=True)

    # After worker completes, give the observer a moment to process events
    await asyncio.sleep(5)

    # Check the final state
    print("Worker completed, checking final state...")
    state_type, message = prefect_core.get_flow_run_state(flow_run.id)
    print(f"Final state after worker run: {state_type} - {message}")

    # Allow both CRASHED and PENDING states - the important thing is
    # that the flow run didn't transition to RUNNING since the pod couldn't start
    acceptable_states = (StateType.CRASHED, StateType.PENDING)
    assert state_type in acceptable_states, (
        f"Expected flow run to be in one of {acceptable_states}, got {state_type}"
    )

    # Collect any events that were generated
    events = []
    with anyio.move_on_after(10):
        while len(events) < 1:
            events = await prefect_core.read_pod_events_for_flow_run(flow_run.id)
            await asyncio.sleep(1)

    async with get_client() as client:
        updated_flow_run = await client.read_flow_run(flow_run.id)
        display.print_flow_run_result(updated_flow_run)

    # Check if we got at least the pending event - but only if we have events
    # It's possible we don't get any events if the pod never started
    if events:
        event_types = {event.event for event in events}
        print(f"Found events: {event_types}")
        assert "prefect.kubernetes.pod.pending" in event_types, (
            f"Expected at least the 'pending' event, got: {event_types}"
        )


@pytest.mark.usefixtures("kind_cluster")
async def test_backoff_limit_exhausted(
    work_pool_name: str,
):
    """Test flow runs with pods that exhaust their backoff limit are marked as crashed."""
    flow_run = await prefect_core.create_flow_run(
        source=DEFAULT_FLOW_SOURCE,
        entrypoint=DEFAULT_FLOW_ENTRYPOINT,
        name="backoff-limit-exhausted",
        work_pool_name=work_pool_name,
        # Use an invalid image to ensure the pod will fail to start
        job_variables=DEFAULT_JOB_VARIABLES | {"backoff_limit": 1},
    )

    display.print_flow_run_created(flow_run)

    with subprocess.Popen(
        ["prefect", "worker", "start", "--pool", work_pool_name],
    ) as worker_process:
        try:
            job = k8s.get_job_for_flow_run(flow_run.name, timeout=120)
            # this loop is a bully
            while job.status and job.status.completion_time is None:
                try:
                    pod_name = k8s.wait_for_pod(job.metadata.name, timeout=15)
                except TimeoutError:
                    break
                # Should hit the backoff limit after final eviction
                k8s.evict_pod(pod_name)
                await asyncio.sleep(1)

            prefect_core.wait_for_flow_run_state(
                flow_run.id, StateType.CRASHED, timeout=60
            )

        finally:
            worker_process.terminate()

    async with get_client() as client:
        updated_flow_run = await client.read_flow_run(flow_run.id)

    assert updated_flow_run.state is not None
    assert updated_flow_run.state.type == StateType.CRASHED

    display.print_flow_run_result(updated_flow_run)

    # Collect events with a more generous timeout
    events = []
    max_events = 0
    with anyio.move_on_after(15):
        while True:
            current_events = await prefect_core.read_pod_events_for_flow_run(
                flow_run.id
            )
            if len(current_events) > max_events:
                max_events = len(current_events)
                events = current_events
                print(
                    f"Found {len(events)} events: {[event.event for event in events]}"
                )
            # If we got at least 5 events, that's enough
            if len(events) >= 5:
                break
            await asyncio.sleep(1)

    # Instead of expecting exactly 6 events, check for at least 5
    assert len(events) >= 5, (
        f"Expected at least 5 events, got {len(events)}: {[event.event for event in events]}"
    )

    # Instead of checking exact order, check the event types
    event_types = {event.event for event in events}
    assert "prefect.kubernetes.pod.pending" in event_types, "Missing pending event"
    assert "prefect.kubernetes.pod.running" in event_types, "Missing running event"
    assert "prefect.kubernetes.pod.failed" in event_types, "Missing failed event"

    # Verify we have events from both pod attempts
    event_list = [event.event for event in events]
    # Count occurrences to verify retries
    pending_count = event_list.count("prefect.kubernetes.pod.pending")
    assert pending_count >= 1, "Expected at least one pending event"
    running_count = event_list.count("prefect.kubernetes.pod.running")
    assert running_count >= 1, "Expected at least one running event"
    failed_count = event_list.count("prefect.kubernetes.pod.failed")
    assert failed_count >= 1, "Expected at least one failed event"

    # Verify the backoff retry happened
    total_events = pending_count + running_count + failed_count
    assert total_events >= 4, (
        f"Expected at least 4 events for retry, got {total_events}"
    )



"""

Module containing the Kubernetes worker used for executing flow runs as Kubernetes jobs.

To start a Kubernetes worker, run the following command:

```bash
prefect worker start --pool 'my-work-pool' --type kubernetes
```

Replace `my-work-pool` with the name of the work pool you want the worker
to poll for flow runs.

### Securing your Prefect Cloud API key
If you are using Prefect Cloud and would like to pass your Prefect Cloud API key to
created jobs via a Kubernetes secret, set the
`PREFECT_INTEGRATIONS_KUBERNETES_WORKER_CREATE_SECRET_FOR_API_KEY` environment variable before
starting your worker:

```bash
export PREFECT_INTEGRATIONS_KUBERNETES_WORKER_CREATE_SECRET_FOR_API_KEY="true"
prefect worker start --pool 'my-work-pool' --type kubernetes
```

Note that your work will need permission to create secrets in the same namespace(s)
that Kubernetes jobs are created in to execute flow runs.

### Using a custom Kubernetes job manifest template

The default template used for Kubernetes job manifests looks like this:
```yaml
---
apiVersion: batch/v1
kind: Job
metadata:
labels: "{{ labels }}"
namespace: "{{ namespace }}"
generateName: "{{ name }}-"
spec:
ttlSecondsAfterFinished: "{{ finished_job_ttl }}"
template:
    spec:
    parallelism: 1
    completions: 1
    restartPolicy: Never
    serviceAccountName: "{{ service_account_name }}"
    containers:
    - name: prefect-job
        env: "{{ env }}"
        image: "{{ image }}"
        imagePullPolicy: "{{ image_pull_policy }}"
        args: "{{ command }}"
```

Each values enclosed in `{{ }}` is a placeholder that will be replaced with
a value at runtime. The values that can be used a placeholders are defined
by the `variables` schema defined in the base job template.

The default job manifest and available variables can be customized on a work pool
by work pool basis. These customizations can be made via the Prefect UI when
creating or editing a work pool.

For example, if you wanted to allow custom memory requests for a Kubernetes work
pool you could update the job manifest template to look like this:

```yaml
---
apiVersion: batch/v1
kind: Job
metadata:
labels: "{{ labels }}"
namespace: "{{ namespace }}"
generateName: "{{ name }}-"
spec:
ttlSecondsAfterFinished: "{{ finished_job_ttl }}"
template:
    spec:
    parallelism: 1
    completions: 1
    restartPolicy: Never
    serviceAccountName: "{{ service_account_name }}"
    containers:
    - name: prefect-job
        env: "{{ env }}"
        image: "{{ image }}"
        imagePullPolicy: "{{ image_pull_policy }}"
        args: "{{ command }}"
        resources:
            requests:
                memory: "{{ memory }}Mi"
            limits:
                memory: 128Mi
```

In this new template, the `memory` placeholder allows customization of the memory
allocated to Kubernetes jobs created by workers in this work pool, but the limit
is hard-coded and cannot be changed by deployments.

For more information about work pools and workers,
checkout out the [Prefect docs](https://docs.prefect.io/concepts/work-pools/).
"""

from __future__ import annotations

import base64
import enum
import json
import shlex
from contextlib import asynccontextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Dict,
    List,
    Optional,
    TypeVar,
    Union,
)

import anyio
import anyio.abc
import kubernetes_asyncio
from jsonpatch import JsonPatch
from kubernetes_asyncio import config
from kubernetes_asyncio.client import (
    ApiClient,
    BatchV1Api,
    CoreV1Api,
    V1Job,
)
from kubernetes_asyncio.client.exceptions import ApiException
from kubernetes_asyncio.client.models import (
    V1ObjectMeta,
    V1Secret,
)
from pydantic import Field, field_validator, model_validator
from tenacity import retry, stop_after_attempt, wait_fixed, wait_random
from typing_extensions import Literal, Self

import prefect
from prefect.client.schemas.objects import Flow as APIFlow
from prefect.exceptions import (
    InfrastructureError,
)
from prefect.utilities.dockerutils import get_prefect_image_name
from prefect.utilities.templating import find_placeholders
from prefect.workers.base import (
    BaseJobConfiguration,
    BaseVariables,
    BaseWorker,
    BaseWorkerResult,
)
from prefect_kubernetes.credentials import KubernetesClusterConfig
from prefect_kubernetes.observer import start_observer, stop_observer
from prefect_kubernetes.settings import KubernetesSettings
from prefect_kubernetes.utilities import (
    KeepAliveClientRequest,
    _slugify_label_key,
    _slugify_label_value,
    _slugify_name,
)

if TYPE_CHECKING:
    from prefect.client.schemas.objects import FlowRun, WorkPool
    from prefect.client.schemas.responses import DeploymentResponse

# Captures flow return type
R = TypeVar("R")


MAX_ATTEMPTS = 3
RETRY_MIN_DELAY_SECONDS = 1
RETRY_MIN_DELAY_JITTER_SECONDS = 0
RETRY_MAX_DELAY_JITTER_SECONDS = 3


def _get_default_job_manifest_template() -> Dict[str, Any]:
    """Returns the default job manifest template used by the Kubernetes worker."""
    return {
        "apiVersion": "batch/v1",
        "kind": "Job",
        "metadata": {
            "labels": "{{ labels }}",
            "namespace": "{{ namespace }}",
            "generateName": "{{ name }}-",
        },
        "spec": {
            "backoffLimit": "{{ backoff_limit }}",
            "ttlSecondsAfterFinished": "{{ finished_job_ttl }}",
            "template": {
                "spec": {
                    "parallelism": 1,
                    "completions": 1,
                    "restartPolicy": "Never",
                    "serviceAccountName": "{{ service_account_name }}",
                    "containers": [
                        {
                            "name": "prefect-job",
                            "env": "{{ env }}",
                            "image": "{{ image }}",
                            "imagePullPolicy": "{{ image_pull_policy }}",
                            "args": "{{ command }}",
                        }
                    ],
                }
            },
        },
    }


def _get_base_job_manifest():
    """Returns a base job manifest to use for manifest validation."""
    return {
        "apiVersion": "batch/v1",
        "kind": "Job",
        "metadata": {"labels": {}},
        "spec": {
            "template": {
                "spec": {
                    "parallelism": 1,
                    "completions": 1,
                    "restartPolicy": "Never",
                    "containers": [
                        {
                            "name": "prefect-job",
                        }
                    ],
                }
            }
        },
    }


class KubernetesImagePullPolicy(enum.Enum):
    """Enum representing the image pull policy options for a Kubernetes job."""

    IF_NOT_PRESENT = "IfNotPresent"
    ALWAYS = "Always"
    NEVER = "Never"


class KubernetesWorkerJobConfiguration(BaseJobConfiguration):
    """
    Configuration class used by the Kubernetes worker.

    An instance of this class is passed to the Kubernetes worker's `run` method
    for each flow run. It contains all of the information necessary to execute
    the flow run as a Kubernetes job.

    Attributes:
        name: The name to give to created Kubernetes job.
        command: The command executed in created Kubernetes jobs to kick off
            flow run execution.
        env: The environment variables to set in created Kubernetes jobs.
        labels: The labels to set on created Kubernetes jobs.
        namespace: The Kubernetes namespace to create Kubernetes jobs in.
        job_manifest: The Kubernetes job manifest to use to create Kubernetes jobs.
        cluster_config: The Kubernetes cluster configuration to use for authentication
            to a Kubernetes cluster.
        job_watch_timeout_seconds: The number of seconds to wait for the job to
            complete before timing out. If `None`, the worker will wait indefinitely.
        pod_watch_timeout_seconds: The number of seconds to wait for the pod to
            complete before timing out.
        stream_output: Whether or not to stream the job's output.
    """

    namespace: str = Field(default="default")
    job_manifest: Dict[str, Any] = Field(
        json_schema_extra=dict(template=_get_default_job_manifest_template())
    )
    cluster_config: Optional[KubernetesClusterConfig] = Field(default=None)
    job_watch_timeout_seconds: Optional[int] = Field(default=None)
    pod_watch_timeout_seconds: int = Field(default=60)
    stream_output: bool = Field(default=True)

    env: Union[Dict[str, Optional[str]], List[Dict[str, Any]]] = Field(
        default_factory=dict
    )

    # internal-use only
    _api_dns_name: Optional[str] = None  # Replaces 'localhost' in API URL

    @model_validator(mode="after")
    def _validate_job_manifest(self) -> Self:
        """
        Validates the job manifest by ensuring the presence of required fields
        and checking for compatible values.
        """
        job_manifest = self.job_manifest
        # Ensure metadata is present
        if "metadata" not in job_manifest:
            job_manifest["metadata"] = {}

        # Ensure labels is present in metadata
        if "labels" not in job_manifest["metadata"]:
            job_manifest["metadata"]["labels"] = {}

        # Ensure namespace is present in metadata
        if "namespace" not in job_manifest["metadata"]:
            job_manifest["metadata"]["namespace"] = self.namespace

        # Check if job includes all required components
        patch = JsonPatch.from_diff(job_manifest, _get_base_job_manifest())
        missing_paths = sorted([op["path"] for op in patch if op["op"] == "add"])
        if missing_paths:
            raise ValueError(
                "Job is missing required attributes at the following paths: "
                f"{', '.join(missing_paths)}"
            )

        # Check if job has compatible values
        incompatible = sorted(
            [
                f"{op['path']} must have value {op['value']!r}"
                for op in patch
                if op["op"] == "replace"
            ]
        )
        if incompatible:
            raise ValueError(
                "Job has incompatible values for the following attributes: "
                f"{', '.join(incompatible)}"
            )

        return self

    @field_validator("env", mode="before")
    @classmethod
    def _coerce_env(cls, v):
        if isinstance(v, list):
            return v
        return {k: str(v) if v is not None else None for k, v in v.items()}

    @staticmethod
    def _base_flow_run_labels(flow_run: "FlowRun") -> Dict[str, str]:
        """
        Generate a dictionary of labels for a flow run job.
        """
        return {
            "prefect.io/flow-run-id": str(flow_run.id),
            "prefect.io/flow-run-name": flow_run.name,
            "prefect.io/version": _slugify_label_value(
                prefect.__version__.split("+")[0]
            ),
        }

    def prepare_for_flow_run(
        self,
        flow_run: "FlowRun",
        deployment: "DeploymentResponse | None" = None,
        flow: "APIFlow | None" = None,
        work_pool: "WorkPool | None" = None,
        worker_name: str | None = None,
    ):
        """
        Prepares the job configuration for a flow run.

        Ensures that necessary values are present in the job manifest and that the
        job manifest is valid.

        Args:
            flow_run: The flow run to prepare the job configuration for
            deployment: The deployment associated with the flow run used for
                preparation.
            flow: The flow associated with the flow run used for preparation.
            work_pool: The work pool associated with the flow run used for preparation.
            worker_name: The name of the worker used for preparation.
        """
        # Save special Kubernetes env vars (like those with valueFrom)
        special_env_vars = []
        if isinstance(self.env, list):
            special_env_vars = [item for item in self.env if "valueFrom" in item]
            original_env = {}
            for item in self.env:
                if "name" in item and "value" in item:
                    original_env[item["name"]] = item.get("value")
            self.env = original_env

        super().prepare_for_flow_run(flow_run, deployment, flow, work_pool, worker_name)

        self._configure_eviction_handling()
        self._update_prefect_api_url_if_local_server()

        # Restore any special env vars with valueFrom before populating the manifest
        if special_env_vars:
            # Convert dict env back to list format
            env_list = [{"name": k, "value": v} for k, v in self.env.items()]
            # Add special env vars back in
            env_list.extend(special_env_vars)
            self.env = env_list

        self._populate_env_in_manifest()
        self._slugify_labels()
        self._populate_image_if_not_present()
        self._populate_command_if_not_present()
        self._populate_generate_name_if_not_present()
        self._propagate_labels_to_pod()

    def _configure_eviction_handling(self):
        """
        Configures eviction handling for the job pod. Needs to run before

        If `backoffLimit` is set to 0, we'll tell the Runner to reschedule
        its flow run when it receives a SIGTERM.

        If `backoffLimit` is set to a positive number, we'll ensure that the
        reschedule SIGTERM handling is not set. Having both a `backoffLimit` and
        reschedule handling set can cause duplicate flow run execution.
        """
        # If backoffLimit is set to 0, we'll tell the Runner to reschedule
        # its flow run when it receives a SIGTERM.
        if self.job_manifest["spec"].get("backoffLimit") == 0:
            if isinstance(self.env, dict):
                self.env["PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR"] = "reschedule"
            elif not any(
                v.get("name") == "PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR"
                for v in self.env
            ):
                self.env.append(
                    {
                        "name": "PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR",
                        "value": "reschedule",
                    }
                )
        # Otherwise, we'll ensure that the reschedule SIGTERM handling is not set.
        else:
            if isinstance(self.env, dict):
                self.env.pop("PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR", None)
            elif any(
                v.get("name") == "PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR"
                for v in self.env
            ):
                self.env = [
                    v
                    for v in self.env
                    if v.get("name") != "PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR"
                ]

    def _populate_env_in_manifest(self):
        """
        Populates environment variables in the job manifest.

        When `env` is templated as a variable in the job manifest it comes in as a
        dictionary. We need to convert it to a list of dictionaries to conform to the
        Kubernetes job manifest schema.

        This function also handles the case where the user has removed the `{{ env }}`
        placeholder and hard coded a value for `env`. In this case, we need to prepend
        our environment variables to the list to ensure Prefect setting propagation.
        An example reason the a user would remove the `{{ env }}` placeholder to
        hardcode Kubernetes secrets in the base job template.
        """
        # Handle both dictionary and list formats for environment variables
        if isinstance(self.env, dict):
            transformed_env = [{"name": k, "value": v} for k, v in self.env.items()]
        else:
            # If env is already a list (k8s format), use it directly
            transformed_env = self.env

        template_env = self.job_manifest["spec"]["template"]["spec"]["containers"][
            0
        ].get("env")

        # If user has removed `{{ env }}` placeholder and hard coded a value for `env`,
        # we need to prepend our environment variables to the list to ensure Prefect
        # setting propagation.
        if isinstance(template_env, list):
            self.job_manifest["spec"]["template"]["spec"]["containers"][0]["env"] = [
                *transformed_env,
                *template_env,
            ]
        # Current templating adds `env` as a dict when the kubernetes manifest requires
        # a list of dicts. Might be able to improve this in the future with a better
        # default `env` value and better typing.
        else:
            self.job_manifest["spec"]["template"]["spec"]["containers"][0]["env"] = (
                transformed_env
            )

    def _update_prefect_api_url_if_local_server(self):
        """If the API URL has been set by the base environment rather than the by the
        user, update the value to ensure connectivity when using a bridge network by
        updating local connections to use the internal host
        """
        if isinstance(self.env, dict):
            if (api_url := self.env.get("PREFECT_API_URL")) and self._api_dns_name:
                self.env["PREFECT_API_URL"] = api_url.replace(
                    "localhost", self._api_dns_name
                ).replace("127.0.0.1", self._api_dns_name)
        else:
            # Handle list format
            for env_var in self.env:
                if (
                    env_var.get("name") == "PREFECT_API_URL"
                    and self._api_dns_name
                    and (value := env_var.get("value"))
                ):
                    env_var["value"] = value.replace(
                        "localhost", self._api_dns_name
                    ).replace("127.0.0.1", self._api_dns_name)

    def _slugify_labels(self):
        """Slugifies the labels in the job manifest."""
        all_labels = {**self.job_manifest["metadata"].get("labels", {}), **self.labels}
        self.job_manifest["metadata"]["labels"] = {
            _slugify_label_key(k): _slugify_label_value(v)
            for k, v in all_labels.items()
        }

    def _populate_image_if_not_present(self):
        """Ensures that the image is present in the job manifest. Populates the image
        with the default Prefect image if it is not present."""
        try:
            if (
                "image"
                not in self.job_manifest["spec"]["template"]["spec"]["containers"][0]
            ):
                self.job_manifest["spec"]["template"]["spec"]["containers"][0][
                    "image"
                ] = get_prefect_image_name()
        except KeyError:
            raise ValueError(
                "Unable to verify image due to invalid job manifest template."
            )

    def _populate_command_if_not_present(self):
        """
        Ensures that the command is present in the job manifest. Populates the command
        with the `prefect -m prefect.engine` if a command is not present.
        """
        try:
            command = self.job_manifest["spec"]["template"]["spec"]["containers"][
                0
            ].get("args")
            if command is None:
                self.job_manifest["spec"]["template"]["spec"]["containers"][0][
                    "args"
                ] = shlex.split(self._base_flow_run_command())
            elif isinstance(command, str):
                self.job_manifest["spec"]["template"]["spec"]["containers"][0][
                    "args"
                ] = shlex.split(command)
            elif not isinstance(command, list):
                raise ValueError(
                    "Invalid job manifest template: 'command' must be a string or list."
                )
        except KeyError:
            raise ValueError(
                "Unable to verify command due to invalid job manifest template."
            )

    def _populate_generate_name_if_not_present(self):
        """Ensures that the generateName is present in the job manifest."""
        manifest_generate_name = self.job_manifest["metadata"].get("generateName", "")
        has_placeholder = len(find_placeholders(manifest_generate_name)) > 0
        # if name wasn't present during template rendering, generateName will be
        # just a hyphen

        manifest_generate_name_templated_with_empty_string = (
            manifest_generate_name == "-"
        )
        if (
            not manifest_generate_name
            or has_placeholder
            or manifest_generate_name_templated_with_empty_string
            or manifest_generate_name == "None-"
        ):
            generate_name = None
            if self.name:
                generate_name = _slugify_name(self.name)
            # _slugify_name will return None if the slugified name in an exception
            if not generate_name:
                generate_name = "prefect-job"
            self.job_manifest["metadata"]["generateName"] = f"{generate_name}-"

    def _propagate_labels_to_pod(self):
        """Propagates Prefect-specific labels to the pod in the job manifest."""
        current_pod_metadata = self.job_manifest["spec"]["template"].get("metadata", {})
        current_pod_labels = current_pod_metadata.get("labels", {})
        all_labels = {**current_pod_labels, **self.labels}

        current_pod_metadata["labels"] = {
            _slugify_label_key(k): _slugify_label_value(v)
            for k, v in all_labels.items()
        }
        self.job_manifest["spec"]["template"]["metadata"] = current_pod_metadata


class KubernetesWorkerVariables(BaseVariables):
    """
    Default variables for the Kubernetes worker.

    The schema for this class is used to populate the `variables` section of the default
    base job template.
    """

    namespace: str = Field(
        default="default", description="The Kubernetes namespace to create jobs within."
    )
    image: Optional[str] = Field(
        default=None,
        description="The image reference of a container image to use for created jobs. "
        "If not set, the latest Prefect image will be used.",
        examples=["docker.io/prefecthq/prefect:3-latest"],
    )
    service_account_name: Optional[str] = Field(
        default=None,
        description="The Kubernetes service account to use for job creation.",
    )
    image_pull_policy: Literal["IfNotPresent", "Always", "Never"] = Field(
        default=KubernetesImagePullPolicy.IF_NOT_PRESENT,
        description="The Kubernetes image pull policy to use for job containers.",
    )
    backoff_limit: int = Field(
        default=0,
        ge=0,
        title="Backoff Limit",
        description=(
            "The number of times Kubernetes will retry a job after pod eviction. "
            "If set to 0, Prefect will reschedule the flow run when the pod is evicted."
        ),
    )
    finished_job_ttl: Optional[int] = Field(
        default=None,
        title="Finished Job TTL",
        description="The number of seconds to retain jobs after completion. If set, "
        "finished jobs will be cleaned up by Kubernetes after the given delay. If not "
        "set, jobs will be retained indefinitely.",
    )
    job_watch_timeout_seconds: Optional[int] = Field(
        default=None,
        description=(
            "Number of seconds to wait for each event emitted by a job before "
            "timing out. If not set, the worker will wait for each event indefinitely."
        ),
    )
    pod_watch_timeout_seconds: int = Field(
        default=60,
        description="Number of seconds to watch for pod creation before timing out.",
    )
    stream_output: bool = Field(
        default=True,
        description=(
            "If set, output will be streamed from the job to local standard output."
        ),
    )
    cluster_config: Optional[KubernetesClusterConfig] = Field(
        default=None,
        description="The Kubernetes cluster config to use for job creation.",
    )


class KubernetesWorkerResult(BaseWorkerResult):
    """Contains information about the final state of a completed process"""


class KubernetesWorker(
    BaseWorker[
        "KubernetesWorkerJobConfiguration",
        "KubernetesWorkerVariables",
        "KubernetesWorkerResult",
    ]
):
    """Prefect worker that executes flow runs within Kubernetes Jobs."""

    type: str = "kubernetes"
    job_configuration = KubernetesWorkerJobConfiguration
    job_configuration_variables = KubernetesWorkerVariables
    _description = (
        "Execute flow runs within jobs scheduled on a Kubernetes cluster. Requires a "
        "Kubernetes cluster."
    )
    _display_name = "Kubernetes"
    _documentation_url = "https://docs.prefect.io/integrations/prefect-kubernetes"
    _logo_url = "https://cdn.sanity.io/images/3ugk85nk/production/2d0b896006ad463b49c28aaac14f31e00e32cfab-250x250.png"  # noqa

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._created_secrets: dict[
            tuple[str, str], KubernetesWorkerJobConfiguration
        ] = {}

    async def _initiate_run(
        self,
        flow_run: "FlowRun",
        configuration: KubernetesWorkerJobConfiguration,
    ) -> None:
        """
        Creates a Kubernetes job to start flow run execution. This method does not
        wait for the job to complete.

        Args:
            flow_run: The flow run to execute
            configuration: The configuration to use when executing the flow run
            task_status: The task status object for the current flow run. If provided,
                the task will be marked as started.
        """
        logger = self.get_flow_run_logger(flow_run)
        async with self._get_configured_kubernetes_client(configuration) as client:
            logger.info("Creating Kubernetes job...")

            await self._create_job(configuration, client)

    async def run(
        self,
        flow_run: "FlowRun",
        configuration: KubernetesWorkerJobConfiguration,
        task_status: anyio.abc.TaskStatus[int] | None = None,
    ) -> KubernetesWorkerResult:
        """
        Executes a flow run within a Kubernetes Job and waits for the flow run
        to complete.

        Args:
            flow_run: The flow run to execute
            configuration: The configuration to use when executing the flow run.
            task_status: The task status object for the current flow run. If provided,
                the task will be marked as started.

        Returns:
            KubernetesWorkerResult: A result object containing information about the
                final state of the flow run
        """
        logger = self.get_flow_run_logger(flow_run)
        async with self._get_configured_kubernetes_client(configuration) as client:
            logger.info("Creating Kubernetes job...")

            job = await self._create_job(configuration, client)

            assert job, "Job should be created"
            pid = f"{job.metadata.namespace}:{job.metadata.name}"
            # Indicate that the job has started
            if task_status is not None:
                task_status.started(pid)

            return KubernetesWorkerResult(identifier=pid, status_code=0)

    async def teardown(self, *exc_info: Any):
        await super().teardown(*exc_info)

        await self._clean_up_created_secrets()

    async def _clean_up_created_secrets(self):
        """Deletes any secrets created during the worker's operation."""
        for key, configuration in self._created_secrets.items():
            async with self._get_configured_kubernetes_client(configuration) as client:
                v1 = CoreV1Api(client)
                result = await v1.delete_namespaced_secret(
                    name=key[0],
                    namespace=key[1],
                )

                if isinstance(result, Exception):
                    self._logger.warning(
                        "Failed to delete created secret with exception: %s", result
                    )

    @asynccontextmanager
    async def _get_configured_kubernetes_client(
        self, configuration: KubernetesWorkerJobConfiguration
    ) -> AsyncGenerator["ApiClient", None]:
        """
        Returns a configured Kubernetes client.
        """
        client = None
        settings = KubernetesSettings()

        if configuration.cluster_config:
            config_dict = configuration.cluster_config.config
            context = configuration.cluster_config.context_name
            client = await config.new_client_from_config_dict(
                config_dict=config_dict,
                context=context,
            )
        else:
            # Try to load in-cluster configuration
            try:
                config.load_incluster_config()
                client = ApiClient()
            except config.ConfigException:
                # If in-cluster config fails, load the local kubeconfig
                client = await config.new_client_from_config()

        if settings.worker.add_tcp_keepalive:
            client.rest_client.pool_manager._request_class = KeepAliveClientRequest

        try:
            yield client
        finally:
            await client.close()

    async def _replace_api_key_with_secret(
        self,
        configuration: KubernetesWorkerJobConfiguration,
        client: "ApiClient",
        secret_name: str | None = None,
        secret_key: str | None = None,
    ):
        """Replaces the PREFECT_API_KEY environment variable with a Kubernetes secret"""
        manifest_env = configuration.job_manifest["spec"]["template"]["spec"][
            "containers"
        ][0].get("env")
        manifest_api_key_env = next(
            (
                env_entry
                for env_entry in manifest_env
                if env_entry.get("name") == "PREFECT_API_KEY"
            ),
            {},
        )
        api_key = manifest_api_key_env.get("value")
        if api_key and not secret_name:
            secret_name = f"prefect-{_slugify_name(self.name)}-api-key"
            secret = await self._upsert_secret(
                name=secret_name,
                value=api_key,
                namespace=configuration.namespace,
                client=client,
            )
            # Store configuration so that we can delete the secret when the worker shuts
            # down
            self._created_secrets[(secret.metadata.name, secret.metadata.namespace)] = (
                configuration
            )
        if secret_name:
            if not secret_key:
                secret_key = "value"
            new_api_env_entry = {
                "name": "PREFECT_API_KEY",
                "valueFrom": {"secretKeyRef": {"name": secret_name, "key": secret_key}},
            }
            manifest_env = [
                entry if entry.get("name") != "PREFECT_API_KEY" else new_api_env_entry
                for entry in manifest_env
            ]
            configuration.job_manifest["spec"]["template"]["spec"]["containers"][0][
                "env"
            ] = manifest_env

    @retry(
        stop=stop_after_attempt(MAX_ATTEMPTS),
        wait=wait_fixed(RETRY_MIN_DELAY_SECONDS)
        + wait_random(
            RETRY_MIN_DELAY_JITTER_SECONDS,
            RETRY_MAX_DELAY_JITTER_SECONDS,
        ),
        reraise=True,
    )
    async def _create_job(
        self, configuration: KubernetesWorkerJobConfiguration, client: "ApiClient"
    ) -> "V1Job":
        """
        Creates a Kubernetes job from a job manifest.
        """
        settings = KubernetesSettings()
        if settings.worker.api_key_secret_name:
            await self._replace_api_key_with_secret(
                configuration=configuration,
                client=client,
                secret_name=settings.worker.api_key_secret_name,
                secret_key=settings.worker.api_key_secret_key,
            )
        elif settings.worker.create_secret_for_api_key:
            await self._replace_api_key_with_secret(
                configuration=configuration, client=client
            )

        try:
            batch_client = BatchV1Api(client)
            job = await batch_client.create_namespaced_job(
                configuration.namespace,
                configuration.job_manifest,
            )
        except kubernetes_asyncio.client.exceptions.ApiException as exc:
            # Parse the reason and message from the response if feasible
            message = ""
            if exc.reason:
                message += ": " + exc.reason
            if exc.body and "message" in (body := json.loads(exc.body)):
                message += ": " + body["message"]

            raise InfrastructureError(
                f"Unable to create Kubernetes job{message}"
            ) from exc

        return job

    async def _upsert_secret(
        self, name: str, value: str, namespace: str, client: "ApiClient"
    ):
        encoded_value = base64.b64encode(value.encode("utf-8")).decode("utf-8")
        core_client = CoreV1Api(client)
        try:
            # Get the current version of the Secret and update it with the
            # new value
            current_secret = await core_client.read_namespaced_secret(
                name=name, namespace=namespace
            )
            current_secret.data = {"value": encoded_value}
            secret = await core_client.replace_namespaced_secret(
                name=name, namespace=namespace, body=current_secret
            )
        except ApiException as exc:
            if exc.status != 404:
                raise
            # Create the secret if it doesn't already exist
            metadata = V1ObjectMeta(name=name, namespace=namespace)
            secret = V1Secret(
                api_version="v1",
                kind="Secret",
                metadata=metadata,
                data={"value": encoded_value},
            )
            secret = await core_client.create_namespaced_secret(
                namespace=namespace, body=secret
            )
        return secret

    @asynccontextmanager
    async def _get_batch_client(
        self, client: "ApiClient"
    ) -> AsyncGenerator["BatchV1Api", None]:
        """
        Context manager for retrieving a Kubernetes batch client.
        """
        try:
            yield BatchV1Api(api_client=client)
        finally:
            await client.close()

    async def __aenter__(self):
        start_observer()
        return await super().__aenter__()

    async def __aexit__(self, *exc_info: Any):
        try:
            await super().__aexit__(*exc_info)
        finally:
            # Need to run after the runs task group exits
            stop_observer()



import subprocess
from typing import Generator

import pytest

from prefect_kubernetes_integration_tests.utils import k8s


@pytest.fixture(scope="session")
def kind_cluster(request: pytest.FixtureRequest) -> Generator[str, None, None]:
    """Create and manage a kind cluster for the test session."""
    cluster_name = request.config.getoption("--cluster-name", default="prefect-test")
    assert isinstance(cluster_name, str)
    k8s.ensure_kind_cluster(cluster_name)
    yield cluster_name


@pytest.fixture(scope="session")
def work_pool_name(
    request: pytest.FixtureRequest, worker_id: str
) -> Generator[str, None, None]:
    """Get the work pool name to use for tests."""
    default_work_pool_name = (
        f"k8s-test-{worker_id}" if worker_id != "master" else "k8s-test"
    )
    work_pool_name = request.config.getoption("--work-pool-name")
    if not isinstance(work_pool_name, str):
        work_pool_name = default_work_pool_name

    subprocess.check_call(
        [
            "prefect",
            "work-pool",
            "create",
            work_pool_name,
            "--type",
            "kubernetes",
            "--overwrite",
        ]
    )

    yield work_pool_name

    subprocess.check_call(
        [
            "prefect",
            "--no-prompt",
            "work-pool",
            "delete",
            work_pool_name,
        ]
    )


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add kubernetes-specific command line options."""
    parser.addoption(
        "--cluster-name",
        action="store",
        default="prefect-test",
        help="Name of the kind cluster to use",
    )
    parser.addoption(
        "--work-pool-name",
        action="store",
        help="Name of the work pool to use",
    )


@pytest.fixture(autouse=True, scope="session")
def _check_k8s_deps():  # pyright: ignore[reportUnusedFunction]
    """Ensure required kubernetes tools are available."""
    k8s.ensure_evict_plugin()


@pytest.fixture(autouse=True, scope="session")
def clean_up_pods() -> Generator[None, None, None]:
    """Clean up pods after tests."""
    yield
    k8s.delete_pods()



from pathlib import Path

import pytest
from kubernetes_asyncio.client.exceptions import ApiValueError
from kubernetes_asyncio.client.models import V1Job
from prefect_kubernetes.jobs import (
    KubernetesJob,
    create_namespaced_job,
    delete_namespaced_job,
    list_namespaced_job,
    patch_namespaced_job,
    read_namespaced_job,
    read_namespaced_job_status,
    replace_namespaced_job,
)


async def test_null_body_raises_error(kubernetes_credentials):
    with pytest.raises(ApiValueError):
        await create_namespaced_job.fn(
            new_job=None, kubernetes_credentials=kubernetes_credentials
        )
    with pytest.raises(ApiValueError):
        await patch_namespaced_job.fn(
            job_updates=None, job_name="", kubernetes_credentials=kubernetes_credentials
        )
    with pytest.raises(ApiValueError):
        await replace_namespaced_job.fn(
            new_job=None, job_name="", kubernetes_credentials=kubernetes_credentials
        )


async def test_create_namespaced_job(kubernetes_credentials, _mock_api_batch_client):
    await create_namespaced_job.fn(
        new_job=V1Job(metadata={"name": "test-job"}),
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )

    assert _mock_api_batch_client.return_value.create_namespaced_job.call_args[1][
        "body"
    ].metadata == {"name": "test-job"}
    assert (
        _mock_api_batch_client.return_value.create_namespaced_job.call_args[1]["a"]
        == "test"
    )


async def test_delete_namespaced_job(kubernetes_credentials, _mock_api_batch_client):
    await delete_namespaced_job.fn(
        job_name="test-job",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )
    assert (
        _mock_api_batch_client.return_value.delete_namespaced_job.call_args[1]["name"]
        == "test-job"
    )
    assert (
        _mock_api_batch_client.return_value.delete_namespaced_job.call_args[1]["a"]
        == "test"
    )


async def test_list_namespaced_job(kubernetes_credentials, _mock_api_batch_client):
    await list_namespaced_job.fn(
        namespace="ns",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )
    assert (
        _mock_api_batch_client.return_value.list_namespaced_job.call_args[1][
            "namespace"
        ]
        == "ns"
    )
    assert (
        _mock_api_batch_client.return_value.list_namespaced_job.call_args[1]["a"]
        == "test"
    )


async def test_patch_namespaced_job(kubernetes_credentials, _mock_api_batch_client):
    await patch_namespaced_job.fn(
        job_updates=V1Job(metadata={"name": "test-job"}),
        job_name="test-job",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )
    assert _mock_api_batch_client.return_value.patch_namespaced_job.call_args[1][
        "body"
    ].metadata == {"name": "test-job"}
    assert (
        _mock_api_batch_client.return_value.patch_namespaced_job.call_args[1]["name"]
        == "test-job"
    )
    assert (
        _mock_api_batch_client.return_value.patch_namespaced_job.call_args[1]["a"]
        == "test"
    )


async def test_read_namespaced_job(kubernetes_credentials, _mock_api_batch_client):
    await read_namespaced_job.fn(
        job_name="test-job",
        namespace="ns",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )
    assert (
        _mock_api_batch_client.return_value.read_namespaced_job.call_args[1]["name"]
        == "test-job"
    )
    assert (
        _mock_api_batch_client.return_value.read_namespaced_job.call_args[1][
            "namespace"
        ]
        == "ns"
    )
    assert (
        _mock_api_batch_client.return_value.read_namespaced_job.call_args[1]["a"]
        == "test"
    )


async def test_replace_namespaced_job(kubernetes_credentials, _mock_api_batch_client):
    await replace_namespaced_job.fn(
        job_name="test-job",
        new_job=V1Job(metadata={"name": "test-job"}),
        namespace="ns",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )
    assert (
        _mock_api_batch_client.return_value.replace_namespaced_job.call_args[1]["name"]
        == "test-job"
    )
    assert (
        _mock_api_batch_client.return_value.replace_namespaced_job.call_args[1][
            "namespace"
        ]
        == "ns"
    )
    assert _mock_api_batch_client.return_value.replace_namespaced_job.call_args[1][
        "body"
    ].metadata == {"name": "test-job"}
    assert (
        _mock_api_batch_client.return_value.replace_namespaced_job.call_args[1]["a"]
        == "test"
    )


async def test_read_namespaced_job_status(
    kubernetes_credentials, _mock_api_batch_client
):
    await read_namespaced_job_status.fn(
        job_name="test-job",
        namespace="ns",
        a="test",
        kubernetes_credentials=kubernetes_credentials,
    )
    assert (
        _mock_api_batch_client.return_value.read_namespaced_job_status.call_args[1][
            "name"
        ]
        == "test-job"
    )
    assert (
        _mock_api_batch_client.return_value.read_namespaced_job_status.call_args[1][
            "namespace"
        ]
        == "ns"
    )
    assert (
        _mock_api_batch_client.return_value.read_namespaced_job_status.call_args[1]["a"]
        == "test"
    )


async def test_job_block_from_job_yaml(kubernetes_credentials):
    DIR = (
        (
            Path.cwd()
            / "src"
            / "integrations"
            / "prefect-kubernetes"
            / "tests"
            / "sample_k8s_resources"
            / "sample_job.yaml"
        )
        if Path.cwd().name == "prefect"
        else Path.cwd() / "tests" / "sample_k8s_resources" / "sample_job.yaml"
    )
    job = KubernetesJob.from_yaml_file(
        credentials=kubernetes_credentials,
        manifest_path=DIR,
    )
    assert isinstance(job, KubernetesJob)
    assert job.v1_job["metadata"]["name"] == "pi"


async def test_job_block_wait_never_called_raises(
    valid_kubernetes_job_block,
    mock_create_namespaced_job,
    mock_delete_namespaced_job,
):
    job_run = await valid_kubernetes_job_block.trigger()

    with pytest.raises(
        ValueError, match="The Kubernetes Job run is not in a completed state"
    ):
        await job_run.fetch_result()


async def test_job_block_generates_default_credentials():
    job_block = KubernetesJob(v1_job=dict(metadata=dict(name="test-job")))
    assert job_block.credentials is not None
    assert job_block.credentials.cluster_config is None



from . import _version
from prefect_kubernetes.credentials import (
    KubernetesCredentials,
    KubernetesClusterConfig,
)  # noqa F401
from prefect_kubernetes.flows import run_namespaced_job  # noqa F401
from prefect_kubernetes.jobs import KubernetesJob  # noqa F401
from prefect_kubernetes.worker import KubernetesWorker  # noqa F401


__version__ = _version.__version__



import asyncio
import os
import subprocess
from typing import Any

import anyio
import pytest

from prefect import get_client
from prefect.states import StateType
from prefect_kubernetes_integration_tests.utils import display, prefect_core

DEFAULT_JOB_VARIABLES: dict[str, Any] = {
    "image": "prefecthq/prefect:3.2.11-python3.12",
}
if os.environ.get("CI", False):
    DEFAULT_JOB_VARIABLES["env"] = {"PREFECT_API_URL": "http://172.17.0.1:4200/api"}

DEFAULT_PARAMETERS = {"n": 1}
# Default source is a simple flow that sleeps
DEFAULT_FLOW_SOURCE = "https://gist.github.com/772d095672484b76da40a4e6158187f0.git"
DEFAULT_FLOW_ENTRYPOINT = "sleeping.py:sleepy"


@pytest.mark.usefixtures("kind_cluster")
async def test_filter_by_namespace(
    work_pool_name: str, monkeypatch: pytest.MonkeyPatch
):
    """Test that the observer doesn't emit events for flow runs in namespaces that are not watched."""
    monkeypatch.setenv(
        "PREFECT_INTEGRATIONS_KUBERNETES_OBSERVER_NAMESPACES",
        "other-namespace,yet-another-namespace",
    )
    flow_run = await prefect_core.create_flow_run(
        source=DEFAULT_FLOW_SOURCE,
        entrypoint=DEFAULT_FLOW_ENTRYPOINT,
        name="filter-by-namespace",
        work_pool_name=work_pool_name,
        job_variables=DEFAULT_JOB_VARIABLES,
        parameters=DEFAULT_PARAMETERS,
    )

    display.print_flow_run_created(flow_run)

    with subprocess.Popen(
        ["prefect", "worker", "start", "--pool", work_pool_name],
    ) as worker_process:
        try:
            prefect_core.wait_for_flow_run_state(
                flow_run.id, StateType.COMPLETED, timeout=30
            )

            async with get_client() as client:
                updated_flow_run = await client.read_flow_run(flow_run.id)

            display.print_flow_run_result(updated_flow_run)

            # Collect events while worker is still running
            events = []
            with anyio.move_on_after(30):
                while len(events) < 3:
                    events = await prefect_core.read_pod_events_for_flow_run(
                        flow_run.id
                    )
                    await asyncio.sleep(1)
        finally:
            worker_process.terminate()

    events = await prefect_core.read_pod_events_for_flow_run(flow_run.id)
    assert len(events) == 0, (
        f"Expected 0 events, got {len(events)}: {[event.event for event in events]}"
    )


@pytest.mark.usefixtures("kind_cluster")
async def test_filter_by_label(work_pool_name: str, monkeypatch: pytest.MonkeyPatch):
    """Test that the observer doesn't emit events for flow runs that don't match the label filter."""
    monkeypatch.setenv(
        "PREFECT_INTEGRATIONS_KUBERNETES_OBSERVER_ADDITIONAL_LABEL_FILTERS",
        "prefect.io/deployment-id=not-real-deployment-id",
    )
    flow_run = await prefect_core.create_flow_run(
        source=DEFAULT_FLOW_SOURCE,
        entrypoint=DEFAULT_FLOW_ENTRYPOINT,
        name="filter-by-label",
        work_pool_name=work_pool_name,
        job_variables=DEFAULT_JOB_VARIABLES,
        parameters=DEFAULT_PARAMETERS,
    )

    display.print_flow_run_created(flow_run)

    with subprocess.Popen(
        ["prefect", "worker", "start", "--pool", work_pool_name],
    ) as worker_process:
        try:
            prefect_core.wait_for_flow_run_state(
                flow_run.id, StateType.COMPLETED, timeout=30
            )

            async with get_client() as client:
                updated_flow_run = await client.read_flow_run(flow_run.id)

            display.print_flow_run_result(updated_flow_run)
        finally:
            worker_process.terminate()

    events = await prefect_core.read_pod_events_for_flow_run(flow_run.id)
    assert len(events) == 0, (
        f"Expected 0 events, got {len(events)}: {[event.event for event in events]}"
    )



import json
import os
from pathlib import Path

import pytest
import toml
from prefect_kubernetes.settings import KubernetesSettings


def test_set_values_via_environment_variables(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(
        "PREFECT_INTEGRATIONS_KUBERNETES_WORKER_API_KEY_SECRET_NAME", "test-secret"
    )
    monkeypatch.setenv(
        "PREFECT_INTEGRATIONS_KUBERNETES_WORKER_CREATE_SECRET_FOR_API_KEY", "true"
    )
    monkeypatch.setenv(
        "PREFECT_INTEGRATIONS_KUBERNETES_WORKER_ADD_TCP_KEEPALIVE", "false"
    )
    monkeypatch.setenv(
        "PREFECT_INTEGRATIONS_KUBERNETES_CLUSTER_UID", "test-cluster-uid"
    )

    settings = KubernetesSettings()

    assert settings.worker.api_key_secret_name == "test-secret"
    assert settings.worker.create_secret_for_api_key is True
    assert settings.worker.add_tcp_keepalive is False
    assert settings.cluster_uid == "test-cluster-uid"


def test_set_values_via_dot_env_file(tmp_path: Path):
    dot_env_path = tmp_path / ".env"
    with open(dot_env_path, "w") as f:
        f.write(
            "PREFECT_INTEGRATIONS_KUBERNETES_WORKER_API_KEY_SECRET_NAME=test-secret\n"
            "PREFECT_INTEGRATIONS_KUBERNETES_WORKER_CREATE_SECRET_FOR_API_KEY=true\n"
            "PREFECT_INTEGRATIONS_KUBERNETES_WORKER_ADD_TCP_KEEPALIVE=false\n"
            "PREFECT_INTEGRATIONS_KUBERNETES_CLUSTER_UID=test-cluster-uid\n"
        )

    original_dir = os.getcwd()
    try:
        os.chdir(tmp_path)
        settings = KubernetesSettings()
    finally:
        os.chdir(original_dir)

    assert settings.worker.api_key_secret_name == "test-secret"
    assert settings.worker.create_secret_for_api_key is True
    assert settings.worker.add_tcp_keepalive is False
    assert settings.cluster_uid == "test-cluster-uid"


def test_set_values_via_prefect_toml_file(tmp_path: Path):
    toml_path = tmp_path / "prefect.toml"
    toml_data = {
        "integrations": {
            "kubernetes": {
                "worker": {
                    "api_key_secret_name": "test-secret",
                    "create_secret_for_api_key": True,
                    "add_tcp_keepalive": False,
                },
                "cluster_uid": "test-cluster-uid",
            },
        },
    }
    toml_path.write_text(toml.dumps(toml_data))

    original_dir = os.getcwd()
    try:
        os.chdir(tmp_path)
        settings = KubernetesSettings()
    finally:
        os.chdir(original_dir)

    assert settings.worker.api_key_secret_name == "test-secret"
    assert settings.worker.create_secret_for_api_key is True
    assert settings.worker.add_tcp_keepalive is False
    assert settings.cluster_uid == "test-cluster-uid"


def test_set_values_via_pyproject_toml_file(tmp_path: Path):
    pyproject_toml_path = tmp_path / "pyproject.toml"
    pyproject_toml_data = {
        "tool": {
            "prefect": {
                "integrations": {
                    "kubernetes": {
                        "cluster_uid": "test-cluster-uid",
                        "worker": {
                            "api_key_secret_name": "test-secret",
                            "create_secret_for_api_key": True,
                            "add_tcp_keepalive": False,
                        },
                    },
                },
            },
        },
    }
    pyproject_toml_path.write_text(toml.dumps(pyproject_toml_data))

    original_dir = os.getcwd()
    try:
        os.chdir(tmp_path)
        settings = KubernetesSettings()
    finally:
        os.chdir(original_dir)

    assert settings.worker.api_key_secret_name == "test-secret"
    assert settings.worker.create_secret_for_api_key is True
    assert settings.worker.add_tcp_keepalive is False
    assert settings.cluster_uid == "test-cluster-uid"


class TestObserverSettings:
    @pytest.mark.parametrize(
        "env_var_value",
        [
            "key1=value1,key2=value2",
            json.dumps({"key1": "value1", "key2": "value2"}),
        ],
    )
    def test_additional_label_filters(
        self, monkeypatch: pytest.MonkeyPatch, env_var_value: str
    ):
        monkeypatch.setenv(
            "PREFECT_INTEGRATIONS_KUBERNETES_OBSERVER_ADDITIONAL_LABEL_FILTERS",
            env_var_value,
        )
        settings = KubernetesSettings()
        assert settings.observer.additional_label_filters == {
            "key1": "value1",
            "key2": "value2",
        }

    @pytest.mark.parametrize(
        "env_var_value",
        [
            "namespace1,namespace2",
            json.dumps(["namespace1", "namespace2"]),
        ],
    )
    def test_namespaces(self, monkeypatch: pytest.MonkeyPatch, env_var_value: str):
        monkeypatch.setenv(
            "PREFECT_INTEGRATIONS_KUBERNETES_OBSERVER_NAMESPACES",
            env_var_value,
        )
        settings = KubernetesSettings()
        assert settings.observer.namespaces == {"namespace1", "namespace2"}



from kubernetes_asyncio.client.models import V1DeleteOptions, V1Service
from prefect_kubernetes.services import (
    create_namespaced_service,
    delete_namespaced_service,
    list_namespaced_service,
    patch_namespaced_service,
    read_namespaced_service,
    replace_namespaced_service,
)


async def test_create_namespaced_service(kubernetes_credentials, _mock_api_core_client):
    await create_namespaced_service.fn(
        kubernetes_credentials=kubernetes_credentials,
        new_service=V1Service(metadata={"name": "test-service"}),
        namespace="default",
    )

    assert _mock_api_core_client.return_value.create_namespaced_service.call_count == 1
    assert _mock_api_core_client.return_value.create_namespaced_service.call_args[1][
        "body"
    ].metadata == {"name": "test-service"}
    assert (
        _mock_api_core_client.return_value.create_namespaced_service.call_args[1][
            "namespace"
        ]
        == "default"
    )


async def test_delete_namespaced_service(kubernetes_credentials, _mock_api_core_client):
    await delete_namespaced_service.fn(
        kubernetes_credentials=kubernetes_credentials,
        service_name="test-service",
        delete_options=V1DeleteOptions(grace_period_seconds=42),
        namespace="default",
    )

    assert _mock_api_core_client.return_value.delete_namespaced_service.call_count == 1
    assert (
        _mock_api_core_client.return_value.delete_namespaced_service.call_args[1][
            "name"
        ]
        == "test-service"
    )
    assert (
        _mock_api_core_client.return_value.delete_namespaced_service.call_args[1][
            "body"
        ].grace_period_seconds
        == 42
    )
    assert (
        _mock_api_core_client.return_value.delete_namespaced_service.call_args[1][
            "namespace"
        ]
        == "default"
    )


async def test_list_namespaced_service(kubernetes_credentials, _mock_api_core_client):
    await list_namespaced_service.fn(
        kubernetes_credentials=kubernetes_credentials,
        namespace="default",
    )

    assert _mock_api_core_client.return_value.list_namespaced_service.call_count == 1
    assert (
        _mock_api_core_client.return_value.list_namespaced_service.call_args[1][
            "namespace"
        ]
        == "default"
    )


async def test_patch_namespaced_service(kubernetes_credentials, _mock_api_core_client):
    await patch_namespaced_service.fn(
        kubernetes_credentials=kubernetes_credentials,
        service_name="test-service-old",
        service_updates=V1Service(metadata={"name": "test-service"}),
        namespace="default",
    )

    assert _mock_api_core_client.return_value.patch_namespaced_service.call_count == 1
    assert (
        _mock_api_core_client.return_value.patch_namespaced_service.call_args[1]["name"]
        == "test-service-old"
    )
    assert _mock_api_core_client.return_value.patch_namespaced_service.call_args[1][
        "body"
    ].metadata == {"name": "test-service"}
    assert (
        _mock_api_core_client.return_value.patch_namespaced_service.call_args[1][
            "namespace"
        ]
        == "default"
    )


async def test_read_namespaced_service(kubernetes_credentials, _mock_api_core_client):
    await read_namespaced_service.fn(
        kubernetes_credentials=kubernetes_credentials,
        service_name="test-service",
        namespace="default",
    )

    assert _mock_api_core_client.return_value.read_namespaced_service.call_count == 1
    assert (
        _mock_api_core_client.return_value.read_namespaced_service.call_args[1]["name"]
        == "test-service"
    )
    assert (
        _mock_api_core_client.return_value.read_namespaced_service.call_args[1][
            "namespace"
        ]
        == "default"
    )


async def test_replace_namespaced_service(
    kubernetes_credentials, _mock_api_core_client
):
    await replace_namespaced_service.fn(
        kubernetes_credentials=kubernetes_credentials,
        service_name="test-service",
        new_service=V1Service(metadata={"labels": {"foo": "bar"}}),
        namespace="default",
    )

    assert _mock_api_core_client.return_value.replace_namespaced_service.call_count == 1
    assert (
        _mock_api_core_client.return_value.replace_namespaced_service.call_args[1][
            "name"
        ]
        == "test-service"
    )
    assert _mock_api_core_client.return_value.replace_namespaced_service.call_args[1][
        "body"
    ].metadata == {"labels": {"foo": "bar"}}
    assert (
        _mock_api_core_client.return_value.replace_namespaced_service.call_args[1][
            "namespace"
        ]
        == "default"
    )



import subprocess
import time

from kubernetes import client, config
from kubernetes.client.models.v1_job import V1Job
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

console = Console()
K8S_NAMESPACE = "default"


def init_k8s_client() -> client.CoreV1Api:
    """Initialize the Kubernetes client."""
    config.load_kube_config()
    return client.CoreV1Api()


def ensure_kind_cluster(name: str = "prefect-test") -> None:
    """Ensure a kind cluster is running."""
    try:
        clusters = subprocess.run(
            ["kind", "get", "clusters"],
            check=True,
            capture_output=True,
            text=True,
        ).stdout

        if name not in clusters:
            console.log(f"Creating Kind cluster: {name}")
            subprocess.check_call(["kind", "create", "cluster", "--name", name])
        else:
            console.log(f"Using existing Kind cluster: {name}")

        # Ensure namespace exists
        v1 = init_k8s_client()
        try:
            v1.read_namespace(K8S_NAMESPACE)
        except client.ApiException:
            v1.create_namespace(client.V1Namespace(metadata={"name": K8S_NAMESPACE}))
            console.log(f"Created namespace: {K8S_NAMESPACE}")

    except subprocess.CalledProcessError:
        subprocess.check_call(["kind", "create", "cluster", "--name", name])
        v1 = init_k8s_client()
        v1.create_namespace(client.V1Namespace(metadata={"name": K8S_NAMESPACE}))


def ensure_evict_plugin() -> None:
    """Ensure kubectl-evict-pod plugin is installed."""
    try:
        subprocess.run(
            ["kubectl", "evict-pod", "--help"],
            check=True,
            capture_output=True,
        )
        console.log("[green]kubectl-evict-pod plugin is installed")
    except subprocess.CalledProcessError:
        console.print(
            Panel(
                "[bold red]kubectl-evict-pod plugin is not installed\n"
                "[white]Install it with: [bold]kubectl krew install evict-pod",
                title="Missing Dependency",
            )
        )
        raise RuntimeError(
            "kubectl-evict-pod plugin is not installed - install it with `kubectl krew install evict-pod`"
        )


def get_job_for_flow_run(flow_run_name: str, timeout: int = 60) -> V1Job:
    """Get the job for a flow run."""
    batch_v1 = client.BatchV1Api()
    start_time = time.time()

    with console.status(f"Looking for job for flow run: {flow_run_name}..."):
        while time.time() - start_time < timeout:
            try:
                jobs = batch_v1.list_namespaced_job(K8S_NAMESPACE)
                for job in jobs.items:
                    if job.metadata.name.startswith(flow_run_name):
                        console.log(f"[green]Found job: {job.metadata.name}")
                        return job
            except client.ApiException as e:
                console.log(f"[yellow]Error checking jobs: {e}")

            time.sleep(2)

    raise TimeoutError(
        f"No job found for flow run {flow_run_name} after {timeout} seconds"
    )


def get_job_name(flow_run_name: str, timeout: int = 60) -> str:
    """Get the full job name for a flow run."""
    job = get_job_for_flow_run(flow_run_name, timeout)
    return job.metadata.name


def wait_for_pod(job_name: str, phase: str = "Running", timeout: int = 60) -> str:
    """Wait for a pod matching the job name to be in a given phase."""
    v1 = init_k8s_client()
    start_time = time.time()

    with console.status(f"Waiting for pod for job: {job_name}..."):
        while time.time() - start_time < timeout:
            try:
                pods = v1.list_namespaced_pod(
                    K8S_NAMESPACE,
                    label_selector=f"job-name={job_name}",
                )

                # Look for running pods, newest first
                sorted_pods = sorted(
                    pods.items,
                    key=lambda x: x.metadata.creation_timestamp,
                    reverse=True,
                )

                for pod in sorted_pods:
                    if pod.status.phase == phase:
                        console.log(f"[green]Found pod: {pod.metadata.name}")
                        return pod.metadata.name

            except client.ApiException as e:
                console.log(f"[yellow]Error checking pods: {e}")

            time.sleep(2)

    raise TimeoutError(
        f"No running pod found for job {job_name} after {timeout} seconds"
    )


def evict_pod(pod_name: str) -> None:
    """Evict a specific pod."""
    v1 = init_k8s_client()

    try:
        # Show initial pod state
        pod = v1.read_namespaced_pod(pod_name, K8S_NAMESPACE)

        pod_summary = {
            "Name": pod.metadata.name,
            "Status": pod.status.phase,
            "Node": pod.spec.node_name,
            "Container Status": "Running"
            if pod.status.container_statuses[0].state.running
            else "Not Running",
            "Image": pod.status.container_statuses[0].image,
            "Created": pod.metadata.creation_timestamp,
        }

        table = Table(title="Pod Details Before Eviction")
        table.add_column("Property", style="cyan")
        table.add_column("Value", style="green")

        for key, value in pod_summary.items():
            table.add_row(key, str(value))

        console.print(table)

        # simulate a pod eviction
        with console.status(f"Evicting pod {pod_name}..."):
            subprocess.check_call(
                ["kubectl", "evict-pod", pod_name, "-n", K8S_NAMESPACE]
            )
            console.print(f"[bold green]Pod {pod_name} evicted successfully")

    except (client.ApiException, subprocess.CalledProcessError) as e:
        console.print(f"[bold red]Failed to evict pod: {e}")
        raise


def delete_pods() -> None:
    """Delete all pods in the default namespace."""
    v1 = init_k8s_client()
    v1.delete_collection_namespaced_pod(K8S_NAMESPACE)



"""Module to define tasks for interacting with Kubernetes jobs."""

from asyncio import sleep
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Type, Union

import yaml
from kubernetes_asyncio.client.models import V1DeleteOptions, V1Job, V1JobList, V1Status
from pydantic import Field
from typing_extensions import Self, TypeAlias

from prefect import task
from prefect.blocks.abstract import JobBlock, JobRun
from prefect.utilities.asyncutils import sync_compatible
from prefect_kubernetes.credentials import KubernetesCredentials
from prefect_kubernetes.exceptions import KubernetesJobTimeoutError
from prefect_kubernetes.pods import list_namespaced_pod, read_namespaced_pod_log

KubernetesManifest: TypeAlias = Union[Dict, Path, str]


def _get_pod_selector(v1_job: V1Job) -> str:
    try:
        return (
            f"controller-uid={v1_job.spec.template.metadata.labels['controller-uid']}"
        )
    except (AttributeError, KeyError):
        raise ValueError("Unable to determine pod selector for job. ")


@task
async def create_namespaced_job(
    kubernetes_credentials: KubernetesCredentials,
    new_job: Union[V1Job, Dict[str, Any]],
    namespace: Optional[str] = "default",
    **kube_kwargs: Dict[str, Any],
) -> V1Job:
    """Task for creating a namespaced Kubernetes job.

    Args:
        kubernetes_credentials: `KubernetesCredentials` block
            holding authentication needed to generate the required API client.
        new_job: A Kubernetes `V1Job` specification.
        namespace: The Kubernetes namespace to create this job in.
        **kube_kwargs: Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`).

    Returns:
        A Kubernetes `V1Job` object.

    Example:
        Create a job in the default namespace:
        ```python
        from prefect import flow
        from prefect_kubernetes.credentials import KubernetesCredentials
        from prefect_kubernetes.jobs import create_namespaced_job
        from kubernetes_asyncio.client.models import V1Job

        @flow
        def kubernetes_orchestrator():
            v1_job_metadata = create_namespaced_job(
                new_job=V1Job(metadata={"labels": {"foo": "bar"}}),
                kubernetes_credentials=KubernetesCredentials.load("k8s-creds"),
            )
        ```
    """
    async with kubernetes_credentials.get_client("batch") as batch_v1_client:
        return await batch_v1_client.create_namespaced_job(
            namespace=namespace,
            body=new_job,
            **kube_kwargs,
        )


@task
async def delete_namespaced_job(
    kubernetes_credentials: KubernetesCredentials,
    job_name: str,
    delete_options: Optional[V1DeleteOptions] = None,
    namespace: Optional[str] = "default",
    **kube_kwargs: Dict[str, Any],
) -> V1Status:
    """Task for deleting a namespaced Kubernetes job.

    Args:
        kubernetes_credentials: `KubernetesCredentials` block
            holding authentication needed to generate the required API client.
        job_name: The name of a job to delete.
        delete_options: A Kubernetes `V1DeleteOptions` object.
        namespace: The Kubernetes namespace to delete this job in.
        **kube_kwargs: Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`).


    Returns:
        A Kubernetes `V1Status` object.

    Example:
        Delete "my-job" in the default namespace:
        ```python
        from kubernetes_asyncio.client.models import V1DeleteOptions
        from prefect import flow
        from prefect_kubernetes.credentials import KubernetesCredentials
        from prefect_kubernetes.jobs import delete_namespaced_job

        @flow
        def kubernetes_orchestrator():
            v1_job_status = delete_namespaced_job(
                job_name="my-job",
                kubernetes_credentials=KubernetesCredentials.load("k8s-creds"),
                delete_options=V1DeleteOptions(propagation_policy="Foreground"),
            )
        ```
    """

    async with kubernetes_credentials.get_client("batch") as batch_v1_client:
        return await batch_v1_client.delete_namespaced_job(
            name=job_name,
            body=delete_options,
            namespace=namespace,
            **kube_kwargs,
        )


@task
async def list_namespaced_job(
    kubernetes_credentials: KubernetesCredentials,
    namespace: Optional[str] = "default",
    **kube_kwargs: Dict[str, Any],
) -> V1JobList:
    """Task for listing namespaced Kubernetes jobs.

    Args:
        kubernetes_credentials: `KubernetesCredentials` block
            holding authentication needed to generate the required API client.
        namespace: The Kubernetes namespace to list jobs from.
        **kube_kwargs: Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`).

    Returns:
        A Kubernetes `V1JobList` object.

    Example:
        List jobs in "my-namespace":
        ```python
        from prefect import flow
        from prefect_kubernetes.credentials import KubernetesCredentials
        from prefect_kubernetes.jobs import list_namespaced_job

        @flow
        def kubernetes_orchestrator():
            namespaced_job_list = list_namespaced_job(
                namespace="my-namespace",
                kubernetes_credentials=KubernetesCredentials.load("k8s-creds"),
            )
        ```
    """
    async with kubernetes_credentials.get_client("batch") as batch_v1_client:
        return await batch_v1_client.list_namespaced_job(
            namespace=namespace,
            **kube_kwargs,
        )


@task
async def patch_namespaced_job(
    kubernetes_credentials: KubernetesCredentials,
    job_name: str,
    job_updates: V1Job,
    namespace: Optional[str] = "default",
    **kube_kwargs: Dict[str, Any],
) -> V1Job:
    """Task for patching a namespaced Kubernetes job.

    Args:
        kubernetes_credentials: KubernetesCredentials block
            holding authentication needed to generate the required API client.
        job_name: The name of a job to patch.
        job_updates: A Kubernetes `V1Job` specification.
        namespace: The Kubernetes namespace to patch this job in.
        **kube_kwargs: Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`).

    Raises:
        ValueError: if `job_name` is `None`.

    Returns:
        A Kubernetes `V1Job` object.

    Example:
        Patch "my-job" in the default namespace:
        ```python
        from prefect import flow
        from prefect_kubernetes.credentials import KubernetesCredentials
        from prefect_kubernetes.jobs import patch_namespaced_job

        from kubernetes_asyncio.client.models import V1Job

        @flow
        def kubernetes_orchestrator():
            v1_job_metadata = patch_namespaced_job(
                job_name="my-job",
                job_updates=V1Job(metadata={"labels": {"foo": "bar"}}}),
                kubernetes_credentials=KubernetesCredentials.load("k8s-creds"),
            )
        ```
    """

    async with kubernetes_credentials.get_client("batch") as batch_v1_client:
        return await batch_v1_client.patch_namespaced_job(
            name=job_name,
            namespace=namespace,
            body=job_updates,
            **kube_kwargs,
        )


@task
async def read_namespaced_job(
    kubernetes_credentials: KubernetesCredentials,
    job_name: str,
    namespace: Optional[str] = "default",
    **kube_kwargs: Dict[str, Any],
) -> V1Job:
    """Task for reading a namespaced Kubernetes job.

    Args:
        kubernetes_credentials: `KubernetesCredentials` block
            holding authentication needed to generate the required API client.
        job_name: The name of a job to read.
        namespace: The Kubernetes namespace to read this job in.
        **kube_kwargs: Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`).

    Raises:
        ValueError: if `job_name` is `None`.

    Returns:
        A Kubernetes `V1Job` object.

    Example:
        Read "my-job" in the default namespace:
        ```python
        from prefect import flow
        from prefect_kubernetes.credentials import KubernetesCredentials
        from prefect_kubernetes.jobs import read_namespaced_job

        @flow
        def kubernetes_orchestrator():
            v1_job_metadata = read_namespaced_job(
                job_name="my-job",
                kubernetes_credentials=KubernetesCredentials.load("k8s-creds"),
            )
        ```
    """
    async with kubernetes_credentials.get_client("batch") as batch_v1_client:
        return await batch_v1_client.read_namespaced_job(
            name=job_name,
            namespace=namespace,
            **kube_kwargs,
        )


@task
async def replace_namespaced_job(
    kubernetes_credentials: KubernetesCredentials,
    job_name: str,
    new_job: V1Job,
    namespace: Optional[str] = "default",
    **kube_kwargs: Dict[str, Any],
) -> V1Job:
    """Task for replacing a namespaced Kubernetes job.

    Args:
        kubernetes_credentials: `KubernetesCredentials` block
            holding authentication needed to generate the required API client.
        job_name: The name of a job to replace.
        new_job: A Kubernetes `V1Job` specification.
        namespace: The Kubernetes namespace to replace this job in.
        **kube_kwargs: Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`).

    Returns:
        A Kubernetes `V1Job` object.

    Example:
        Replace "my-job" in the default namespace:
        ```python
        from prefect import flow
        from prefect_kubernetes.credentials import KubernetesCredentials
        from prefect_kubernetes.jobs import replace_namespaced_job

        @flow
        def kubernetes_orchestrator():
            v1_job_metadata = replace_namespaced_job(
                new_job=V1Job(metadata={"labels": {"foo": "bar"}}),
                job_name="my-job",
                kubernetes_credentials=KubernetesCredentials.load("k8s-creds"),
            )
        ```
    """
    async with kubernetes_credentials.get_client("batch") as batch_v1_client:
        return await batch_v1_client.replace_namespaced_job(
            name=job_name,
            body=new_job,
            namespace=namespace,
            **kube_kwargs,
        )


@task
async def read_namespaced_job_status(
    kubernetes_credentials: KubernetesCredentials,
    job_name: str,
    namespace: Optional[str] = "default",
    **kube_kwargs: Dict[str, Any],
) -> V1Job:
    """Task for fetching status of a namespaced Kubernetes job.

    Args:
        kubernetes_credentials: `KubernetesCredentials` block
            holding authentication needed to generate the required API client.
        job_name: The name of a job to fetch status for.
        namespace: The Kubernetes namespace to fetch status of job in.
        **kube_kwargs: Optional extra keyword arguments to pass to the
            Kubernetes API (e.g. `{"pretty": "...", "dry_run": "..."}`).

    Returns:
        A Kubernetes `V1JobStatus` object.

    Example:
        Fetch status of a job in the default namespace:
        ```python
        from prefect import flow
        from prefect_kubernetes.credentials import KubernetesCredentials
        from prefect_kubernetes.jobs import read_namespaced_job_status

        @flow
        def kubernetes_orchestrator():
            v1_job_status = read_namespaced_job_status(
                kubernetes_credentials=KubernetesCredentials.load("k8s-creds"),
                job_name="my-job",
            )
        ```
    """
    async with kubernetes_credentials.get_client("batch") as batch_v1_client:
        return await batch_v1_client.read_namespaced_job_status(
            name=job_name,
            namespace=namespace,
            **kube_kwargs,
        )


class KubernetesJobRun(JobRun[Dict[str, Any]]):
    """A container representing a run of a Kubernetes job."""

    def __init__(
        self,
        kubernetes_job: "KubernetesJob",
        v1_job_model: Union[V1Job, Dict[str, Any]],
    ):
        self.pod_logs = None

        self._completed = False
        self._kubernetes_job = kubernetes_job
        self._v1_job_model = v1_job_model

    @property
    def v1_job_model(self) -> dict[str, Any]:
        return (
            self._v1_job_model
            if isinstance(self._v1_job_model, dict)
            else self._v1_job_model.to_dict()
        )

    async def _cleanup(self):
        """Deletes the Kubernetes job resource."""
        job_name = self._v1_job_model.get("metadata", {}).get("name")

        delete_options = V1DeleteOptions(propagation_policy="Foreground")

        deleted_v1_job = await delete_namespaced_job.fn(
            kubernetes_credentials=self._kubernetes_job.credentials,
            job_name=job_name,
            delete_options=delete_options,
            namespace=self._kubernetes_job.namespace,
            **self._kubernetes_job.api_kwargs,
        )
        self.logger.info(f"Job {job_name} deleted with {deleted_v1_job.status!r}.")

    @sync_compatible
    async def wait_for_completion(self, print_func: Optional[Callable] = None):
        """Waits for the job to complete.

        If the job has `delete_after_completion` set to `True`,
        the job will be deleted if it is observed by this method
        to enter a completed state.

        Raises:
            RuntimeError: If the Kubernetes job fails.
            KubernetesJobTimeoutError: If the Kubernetes job times out.
            ValueError: If `wait_for_completion` is never called.
        """
        self.pod_logs = {}

        elapsed_time = 0

        while not self._completed:
            job_expired = (
                elapsed_time > self._kubernetes_job.timeout_seconds
                if self._kubernetes_job.timeout_seconds
                else False
            )
            if job_expired:
                raise KubernetesJobTimeoutError(
                    f"Job timed out after {elapsed_time} seconds."
                )

            v1_job = await read_namespaced_job.fn(
                kubernetes_credentials=self._kubernetes_job.credentials,
                job_name=self.v1_job_model.get("metadata", {}).get("name"),
                namespace=self._kubernetes_job.namespace,
                **self._kubernetes_job.api_kwargs,
            )
            pod_selector = _get_pod_selector(v1_job)
            v1_pod_list = await list_namespaced_pod.fn(
                kubernetes_credentials=self._kubernetes_job.credentials,
                namespace=self._kubernetes_job.namespace,
                label_selector=pod_selector,
                **self._kubernetes_job.api_kwargs,
            )

            for pod in v1_pod_list.items:
                pod_name = pod.metadata.name

                if pod.status.phase == "Pending" or pod_name in self.pod_logs.keys():
                    continue

                self.logger.info(f"Capturing logs for pod {pod_name!r}.")

                self.pod_logs[pod_name] = await read_namespaced_pod_log.fn(
                    kubernetes_credentials=self._kubernetes_job.credentials,
                    pod_name=pod_name,
                    container=v1_job.spec.template.spec.containers[0].name,
                    namespace=self._kubernetes_job.namespace,
                    print_func=print_func,
                    **self._kubernetes_job.api_kwargs,
                )

            if v1_job.status.active:
                await sleep(self._kubernetes_job.interval_seconds)
                if self._kubernetes_job.timeout_seconds:
                    elapsed_time += self._kubernetes_job.interval_seconds
            elif v1_job.status.conditions:
                final_completed_conditions = [
                    condition.type == "Complete"
                    for condition in v1_job.status.conditions
                    if condition.status == "True"
                ]
                if final_completed_conditions and any(final_completed_conditions):
                    self._completed = True
                    self.logger.info(
                        f"Job {v1_job.metadata.name!r} has "
                        f"completed with {v1_job.status.succeeded} pods."
                    )
                elif final_completed_conditions:
                    failed_conditions = [
                        condition.reason
                        for condition in v1_job.status.conditions
                        if condition.type == "Failed"
                    ]
                    raise RuntimeError(
                        f"Job {v1_job.metadata.name!r} failed due to "
                        f"{failed_conditions}, check the Kubernetes pod logs "
                        f"for more information."
                    )

        if self._kubernetes_job.delete_after_completion:
            await self._cleanup()

    @sync_compatible
    async def fetch_result(self) -> Dict[str, Any]:
        """Fetch the results of the job.

        Returns:
            The logs from each of the pods in the job.

        Raises:
            ValueError: If this method is called when the job has
                a non-terminal state.
        """

        if not self._completed:
            raise ValueError(
                "The Kubernetes Job run is not in a completed state - "
                "be sure to call `wait_for_completion` before attempting "
                "to fetch the result."
            )
        return self.pod_logs


class KubernetesJob(JobBlock):
    """A block representing a Kubernetes job configuration."""

    v1_job: Dict[str, Any] = Field(
        default=...,
        title="Job Manifest",
        description=(
            "The Kubernetes job manifest to run. This dictionary can be produced "
            "using `yaml.safe_load`."
        ),
    )
    api_kwargs: Dict[str, Any] = Field(
        default_factory=dict,
        title="Additional API Arguments",
        description="Additional arguments to include in Kubernetes API calls.",
        examples=[{"pretty": "true"}],
    )
    credentials: KubernetesCredentials = Field(
        default_factory=KubernetesCredentials,
        description="The credentials to configure a client from.",
    )
    delete_after_completion: bool = Field(
        default=True,
        description="Whether to delete the job after it has completed.",
    )
    interval_seconds: int = Field(
        default=5,
        description="The number of seconds to wait between job status checks.",
    )
    namespace: str = Field(
        default="default",
        description="The namespace to create and run the job in.",
    )
    timeout_seconds: Optional[int] = Field(
        default=None,
        description="The number of seconds to wait for the job run before timing out.",
    )

    _block_type_name = "Kubernetes Job"
    _block_type_slug = "k8s-job"
    _logo_url = "https://cdn.sanity.io/images/3ugk85nk/production/2d0b896006ad463b49c28aaac14f31e00e32cfab-250x250.png"  # noqa: E501
    _documentation_url = "https://docs.prefect.io/integrations/prefect-kubernetes"  # noqa

    @sync_compatible
    async def trigger(self):
        """Create a Kubernetes job and return a `KubernetesJobRun` object."""

        await create_namespaced_job.fn(
            kubernetes_credentials=self.credentials,
            new_job=self.v1_job,
            namespace=self.namespace,
            **self.api_kwargs,
        )

        return KubernetesJobRun(kubernetes_job=self, v1_job_model=self.v1_job)

    @classmethod
    def from_yaml_file(
        cls: Type[Self], manifest_path: Union[Path, str], **kwargs
    ) -> Self:
        """Create a `KubernetesJob` from a YAML file.

        Args:
            manifest_path: The YAML file to create the `KubernetesJob` from.

        Returns:
            A KubernetesJob object.
        """
        with open(manifest_path, "r") as yaml_stream:
            yaml_dict = yaml.safe_load(yaml_stream)

        return cls(v1_job=yaml_dict, **kwargs)



import uuid
from contextlib import asynccontextmanager
from time import sleep
from unittest.mock import AsyncMock, MagicMock

import pytest
from prefect_kubernetes.observer import (
    _replicate_pod_event,
    start_observer,
    stop_observer,
)

from prefect.events.schemas.events import RelatedResource, Resource


@pytest.fixture
def mock_events_client(monkeypatch: pytest.MonkeyPatch):
    events_client = AsyncMock()

    @asynccontextmanager
    async def mock_get_events_client():
        try:
            yield events_client
        finally:
            pass

    monkeypatch.setattr(
        "prefect_kubernetes.observer.get_events_client", mock_get_events_client
    )
    monkeypatch.setattr("prefect_kubernetes.observer.events_client", events_client)
    return events_client


@pytest.fixture
def mock_orchestration_client(monkeypatch: pytest.MonkeyPatch):
    orchestration_client = AsyncMock()
    json_response = MagicMock()
    json_response.json.return_value = {"events": [{"id": "existing-event"}]}
    orchestration_client.request.return_value = json_response

    @asynccontextmanager
    async def mock_get_orchestration_client():
        try:
            yield orchestration_client
        finally:
            pass

    monkeypatch.setattr(
        "prefect_kubernetes.observer.get_client",
        mock_get_orchestration_client,
    )
    monkeypatch.setattr(
        "prefect_kubernetes.observer.orchestration_client", orchestration_client
    )
    return orchestration_client


class TestReplicatePodEvent:
    async def test_minimal(self, mock_events_client: AsyncMock):
        flow_run_id = uuid.uuid4()
        pod_id = uuid.uuid4()

        await _replicate_pod_event(
            event={"type": "ADDED", "status": {"phase": "Running"}},
            uid=str(pod_id),
            name="test",
            namespace="test",
            labels={
                "prefect.io/flow-run-id": str(flow_run_id),
                "prefect.io/flow-run-name": "test",
            },
            status={"phase": "Running"},
            logger=MagicMock(),
        )

        emitted_event = mock_events_client.emit.call_args[1]["event"]
        assert emitted_event.event == "prefect.kubernetes.pod.running"
        assert emitted_event.resource == Resource(
            {
                "prefect.resource.id": f"prefect.kubernetes.pod.{pod_id}",
                "prefect.resource.name": "test",
                "kubernetes.namespace": "test",
            }
        )
        assert emitted_event.related == [
            RelatedResource.model_validate(
                {
                    "prefect.resource.id": f"prefect.flow-run.{flow_run_id}",
                    "prefect.resource.role": "flow-run",
                    "prefect.resource.name": "test",
                }
            )
        ]

    async def test_deterministic_event_id(self, mock_events_client: AsyncMock):
        """Test that the event ID is deterministic"""
        pod_id = uuid.uuid4()
        await _replicate_pod_event(
            event={"type": "ADDED", "status": {"phase": "Running"}},
            uid=str(pod_id),
            name="test",
            namespace="test",
            labels={
                "prefect.io/flow-run-id": str(uuid.uuid4()),
                "prefect.io/flow-run-name": "test-run",
            },
            status={"phase": "Running"},
            logger=MagicMock(),
        )

        first_event_id = mock_events_client.emit.call_args[1]["event"].id
        mock_events_client.emit.reset_mock()

        # Call the function again
        await _replicate_pod_event(
            event={"type": "ADDED", "status": {"phase": "Running"}},
            uid=str(pod_id),
            name="test",
            namespace="test",
            labels={
                "prefect.io/flow-run-id": str(uuid.uuid4()),
                "prefect.io/flow-run-name": "test-run",
            },
            status={"phase": "Running"},
            logger=MagicMock(),
        )

        second_event_id = mock_events_client.emit.call_args[1]["event"].id
        assert first_event_id == second_event_id

    async def test_evicted_pod(self, mock_events_client: AsyncMock):
        """Test handling of evicted pods"""
        pod_id = uuid.uuid4()

        await _replicate_pod_event(
            event={"type": "MODIFIED"},
            uid=str(pod_id),
            name="test",
            namespace="test",
            labels={
                "prefect.io/flow-run-id": str(uuid.uuid4()),
                "prefect.io/flow-run-name": "test-run",
            },
            status={
                "phase": "Failed",
                "container_statuses": [
                    {"state": {"terminated": {"reason": "OOMKilled"}}}
                ],
            },
            logger=MagicMock(),
        )

        emitted_event = mock_events_client.emit.call_args[1]["event"]
        assert emitted_event.event == "prefect.kubernetes.pod.evicted"
        assert emitted_event.resource == Resource(
            {
                "prefect.resource.id": f"prefect.kubernetes.pod.{pod_id}",
                "prefect.resource.name": "test",
                "kubernetes.namespace": "test",
                "kubernetes.reason": "OOMKilled",
            },
        )

    async def test_all_related_resources(self, mock_events_client: AsyncMock):
        """Test that all possible related resources are included"""
        flow_run_id = uuid.uuid4()
        deployment_id = uuid.uuid4()
        flow_id = uuid.uuid4()
        work_pool_id = uuid.uuid4()
        pod_id = uuid.uuid4()

        await _replicate_pod_event(
            event={"type": "ADDED"},
            uid=str(pod_id),
            name="test",
            namespace="test",
            labels={
                "prefect.io/flow-run-id": str(flow_run_id),
                "prefect.io/flow-run-name": "test-run",
                "prefect.io/deployment-id": str(deployment_id),
                "prefect.io/deployment-name": "test-deployment",
                "prefect.io/flow-id": str(flow_id),
                "prefect.io/flow-name": "test-flow",
                "prefect.io/work-pool-id": str(work_pool_id),
                "prefect.io/work-pool-name": "test-pool",
                "prefect.io/worker-name": "test-worker",
            },
            status={"phase": "Running"},
            logger=MagicMock(),
        )

        mock_events_client.emit.assert_called_once()
        emitted_event = mock_events_client.emit.call_args[1]["event"]
        related_resources = emitted_event.related

        # Verify all related resources are present
        resource_ids = {
            r.model_dump()["prefect.resource.id"] for r in related_resources
        }
        assert resource_ids == {
            f"prefect.flow-run.{flow_run_id}",
            f"prefect.deployment.{deployment_id}",
            f"prefect.flow.{flow_id}",
            f"prefect.work-pool.{work_pool_id}",
            "prefect.worker.kubernetes.test-worker",
        }

        resource_names = {
            r.model_dump()["prefect.resource.name"] for r in related_resources
        }
        assert resource_names == {
            "test-run",
            "test-deployment",
            "test-flow",
            "test-pool",
            "test-worker",
        }

    @pytest.mark.usefixtures("mock_orchestration_client")
    async def test_event_deduplication(self, mock_events_client: AsyncMock):
        """Test that checks from existing events when receiving events on startup"""
        pod_id = uuid.uuid4()
        await _replicate_pod_event(
            # Event types with None are received when reading current cluster state
            event={"type": None},
            uid=str(pod_id),
            name="test",
            namespace="test",
            labels={"prefect.io/flow-run-id": str(uuid.uuid4())},
            status={"phase": "Running"},
            logger=MagicMock(),
        )

        # Verify no event was emitted since one already existed
        mock_events_client.emit.assert_not_called()

    @pytest.mark.parametrize("phase", ["Pending", "Running", "Succeeded", "Failed"])
    async def test_different_phases(self, mock_events_client: AsyncMock, phase: str):
        """Test handling of different pod phases"""
        pod_id = uuid.uuid4()
        flow_run_id = uuid.uuid4()

        mock_events_client.emit.reset_mock()
        await _replicate_pod_event(
            event={"type": "ADDED"},
            uid=str(pod_id),
            name="test",
            namespace="test",
            labels={
                "prefect.io/flow-run-id": str(flow_run_id),
                "prefect.io/flow-run-name": "test-run",
            },
            status={"phase": phase},
            logger=MagicMock(),
        )

        mock_events_client.emit.assert_called_once()
        emitted_event = mock_events_client.emit.call_args[1]["event"]
        assert emitted_event.event == f"prefect.kubernetes.pod.{phase.lower()}"


class TestStartAndStopObserver:
    @pytest.mark.timeout(10)
    @pytest.mark.usefixtures("mock_events_client", "mock_orchestration_client")
    def test_start_and_stop(self, monkeypatch: pytest.MonkeyPatch):
        """
        Test that the observer can be started and stopped without errors
        and without hanging.
        """
        start_observer()
        sleep(1)
        stop_observer()
